package ru.dbotthepony.kstarbound.defs

import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap

private fun flattenJsonPrimitive(input: JsonPrimitive): Any {
	if (input.isNumber) {
		return input.asNumber
	} else if (input.isString) {
		return input.asString.intern()
	} else {
		return input.asBoolean
	}
}

private fun flattenJsonArray(input: JsonArray): ArrayList<Any> {
	val flattened = ArrayList<Any>(input.size())

	for (v in input) {
		when (v) {
			is JsonObject -> flattened.add(flattenJsonObject(v))
			is JsonArray -> flattened.add(flattenJsonArray(v))
			is JsonPrimitive -> flattened.add(flattenJsonPrimitive(v))
			// is JsonNull -> baked.add(null)
		}
	}

	return flattened
}

private fun flattenJsonObject(input: JsonObject): Object2ObjectArrayMap<String, Any> {
	val flattened = Object2ObjectArrayMap<String, Any>()

	for ((k, v) in input.entrySet()) {
		when (v) {
			is JsonObject -> flattened[k] = flattenJsonObject(v)
			is JsonArray -> flattened[k] = flattenJsonArray(v)
			is JsonPrimitive -> flattened[k] = flattenJsonPrimitive(v)
		}
	}

	return flattened
}

fun flattenJsonElement(input: JsonElement): Any? {
	return when (input) {
		is JsonObject -> flattenJsonObject(input)
		is JsonArray -> flattenJsonArray(input)
		is JsonPrimitive -> flattenJsonPrimitive(input)
		is JsonNull -> null
		else -> throw IllegalArgumentException("Unknown argument $input")
	}
}
