package ru.dbotthepony.kstarbound.defs.tile

import com.google.gson.GsonBuilder
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.io.json.KConcreteTypeAdapter

const val TILE_COLOR_VARIANTS = 9

data class RenderParameters(
	val texture: String,
	val variants: Int = 0,
	val multiColored: Boolean = false,
	val occludesBelow: Boolean = false,
	val lightTransparent: Boolean = false,
	val zLevel: Int,
) {
	val absoluteTexturePath: String

	init {
		val dir = Starbound.readingFolder

		if (dir == null || texture[0] == '/') {
			absoluteTexturePath = texture
		} else {
			absoluteTexturePath = "$dir/$texture"
		}
	}

	companion object {
		val ADAPTER = KConcreteTypeAdapter.Builder(RenderParameters::class)
			.auto(
				RenderParameters::texture,
				RenderParameters::variants,
				RenderParameters::multiColored,
				RenderParameters::occludesBelow,
				RenderParameters::lightTransparent,
				RenderParameters::zLevel,
			)
			.build()

		fun registerGson(gsonBuilder: GsonBuilder) {
			gsonBuilder.registerTypeAdapter(RenderParameters::class.java, ADAPTER)
		}
	}
}
