package ru.dbotthepony.kstarbound.io.json

import com.google.common.collect.ImmutableMap
import com.google.gson.TypeAdapter
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.Starbound

class MapAdapter<K, V>(val keyAdapter: TypeAdapter<K>, val valueAdapter: TypeAdapter<V>) : TypeAdapter<Map<K, V>>() {
	constructor(keyType: Class<K>, valueType: Class<V>) : this(LazyTypeProvider(keyType), LazyTypeProvider(valueType))

	override fun write(out: JsonWriter, value: Map<K, V>) {
		out.beginArray()

		for ((k, v) in value) {
			out.beginArray()
			keyAdapter.write(out, k)
			valueAdapter.write(out, v)
			out.endArray()
		}

		out.endArray()
	}

	override fun read(reader: JsonReader): Map<K, V> {
		reader.beginArray()

		val builder = ImmutableMap.builder<K, V>()

		while (reader.peek() != JsonToken.END_ARRAY) {
			reader.beginArray()
			builder.put(keyAdapter.read(reader), valueAdapter.read(reader))
			reader.endArray()
		}

		reader.endArray()

		return builder.build()
	}
}
