package ru.dbotthepony.kstarbound.world


/**
 * Кортеж чанка, который содержит родителя (мир) и соседей (кортежи чанков)
 */
interface IWorldChunkTuple<WorldType : World<WorldType, ChunkType>, ChunkType : Chunk<WorldType, ChunkType>> {
	val world: WorldType
	val chunk: ChunkType
	val top: IWorldChunkTuple<WorldType, ChunkType>?
	val left: IWorldChunkTuple<WorldType, ChunkType>?
	val right: IWorldChunkTuple<WorldType, ChunkType>?
	val bottom: IWorldChunkTuple<WorldType, ChunkType>?

	val topLeft: IWorldChunkTuple<WorldType, ChunkType>?
	val topRight: IWorldChunkTuple<WorldType, ChunkType>?
	val bottomLeft: IWorldChunkTuple<WorldType, ChunkType>?
	val bottomRight: IWorldChunkTuple<WorldType, ChunkType>?
}

class ProxiedWorldChunkTuple<WorldType : World<WorldType, ChunkType>, ChunkType : Chunk<WorldType, ChunkType>>(
	private val parent: IWorldChunkTuple<WorldType, ChunkType>
) : IWorldChunkTuple<WorldType, ChunkType> {
	override val world get() = parent.world
	override val chunk get() = parent.chunk

	override val top: IWorldChunkTuple<WorldType, ChunkType>? get() = parent.top?.let(::ProxiedWorldChunkTuple)
	override val left: IWorldChunkTuple<WorldType, ChunkType>? get() = parent.left?.let(::ProxiedWorldChunkTuple)
	override val right: IWorldChunkTuple<WorldType, ChunkType>? get() = parent.right?.let(::ProxiedWorldChunkTuple)
	override val bottom: IWorldChunkTuple<WorldType, ChunkType>? get() = parent.bottom?.let(::ProxiedWorldChunkTuple)
	override val topLeft: IWorldChunkTuple<WorldType, ChunkType>? get() = parent.topLeft?.let(::ProxiedWorldChunkTuple)
	override val topRight: IWorldChunkTuple<WorldType, ChunkType>? get() = parent.topRight?.let(::ProxiedWorldChunkTuple)
	override val bottomLeft: IWorldChunkTuple<WorldType, ChunkType>? get() = parent.bottomLeft?.let(::ProxiedWorldChunkTuple)
	override val bottomRight: IWorldChunkTuple<WorldType, ChunkType>? get() = parent.bottomRight?.let(::ProxiedWorldChunkTuple)
}

class InstantWorldChunkTuple<WorldType : World<WorldType, ChunkType>, ChunkType : Chunk<WorldType, ChunkType>>(
	override val world: WorldType,
	override val chunk: ChunkType
) : IWorldChunkTuple<WorldType, ChunkType> {

	private val _top = world[chunk.top]
	private val _left = world[chunk.left]
	private val _right = world[chunk.right]
	private val _bottom = world[chunk.bottom]
	private val _topLeft = world[chunk.topLeft]
	private val _topRight = world[chunk.topRight]
	private val _bottomLeft = world[chunk.bottomLeft]
	private val _bottomRight = world[chunk.bottomRight]

	override val top: IWorldChunkTuple<WorldType, ChunkType>? 					by lazy { _top?.let { InstantWorldChunkTuple(world, it) } }
	override val left: IWorldChunkTuple<WorldType, ChunkType>? 					by lazy { _left?.let { InstantWorldChunkTuple(world, it) } }
	override val right: IWorldChunkTuple<WorldType, ChunkType>? 				by lazy { _right?.let { InstantWorldChunkTuple(world, it) } }
	override val bottom: IWorldChunkTuple<WorldType, ChunkType>? 				by lazy { _bottom?.let { InstantWorldChunkTuple(world, it) } }
	override val topLeft: IWorldChunkTuple<WorldType, ChunkType>? 				by lazy { _topLeft?.let { InstantWorldChunkTuple(world, it) } }
	override val topRight: IWorldChunkTuple<WorldType, ChunkType>? 				by lazy { _topRight?.let { InstantWorldChunkTuple(world, it) } }
	override val bottomLeft: IWorldChunkTuple<WorldType, ChunkType>? 			by lazy { _bottomLeft?.let { InstantWorldChunkTuple(world, it) } }
	override val bottomRight: IWorldChunkTuple<WorldType, ChunkType>? 			by lazy { _bottomRight?.let { InstantWorldChunkTuple(world, it) } }
}
